//! Manipulate pdf documents.

use std::{
    collections::BTreeMap,
    path::{Path, PathBuf},
};

use anyhow::Result;
use lopdf::{Dictionary, Document, Object, ObjectId};

const METADATA_TITLE: &str = "Title";
const METADATA_PRODUCER: &str = "Producer";
const PDF_VERSION: &str = "1.8";
const PRODUCER: &str = "NZZ Downloader";

/// Merge the provided pdfs in the `input` vector to one pdf in `out`, setting its title to
/// `title`.
///
/// The code is from https://github.com/J-F-Liu/lopdf/blob/6b04581640e061bfeb39b585e50a7e9d102b8fe2/examples/merge.rs
/// with some modifications. I have no clue about PDF structure and this is still a bit of a
/// mistery to me.
pub fn merge(input: Vec<PathBuf>, out: &Path, title: &str) -> Result<()> {
    let mut max_id = 1;
    let mut documents_pages = BTreeMap::new();
    let mut documents_objects = BTreeMap::new();
    let mut merged_doc = Document::with_version(PDF_VERSION);

    for pdf in input {
        let mut doc = Document::load(pdf)?;

        doc.renumber_objects_with(max_id);

        max_id = doc.max_id + 1;

        documents_pages.extend(
            doc.get_pages()
                .into_values()
                .map(|object_id| (object_id, doc.get_object(object_id).unwrap().to_owned()))
                .collect::<BTreeMap<ObjectId, Object>>(),
        );
        documents_objects.extend(doc.objects);
    }

    let mut catalog_object: Option<(ObjectId, Object)> = None;
    let mut pages_object: Option<(ObjectId, Object)> = None;

    // Process all objects except "Page" type
    for (object_id, object) in documents_objects.iter() {
        // We have to ignore "Page" (as are processed later), "Outlines" and "Outline" objects
        // All other objects should be collected and inserted into the main Document
        match object.type_name().unwrap_or("") {
            "Catalog" => {
                // Collect a first "Catalog" object and use it for the future "Pages"
                catalog_object = Some((
                    if let Some((id, _)) = catalog_object {
                        id
                    } else {
                        *object_id
                    },
                    object.clone(),
                ));
            }
            "Pages" => {
                // Collect and update a first "Pages" object and use it for the future "Catalog"
                // We have also to merge all dictionaries of the old and the new "Pages" object
                if let Ok(dictionary) = object.as_dict() {
                    let mut dictionary = dictionary.clone();
                    if let Some((_, ref object)) = pages_object {
                        if let Ok(old_dictionary) = object.as_dict() {
                            dictionary.extend(old_dictionary);
                        }
                    }

                    pages_object = Some((
                        if let Some((id, _)) = pages_object {
                            id
                        } else {
                            *object_id
                        },
                        Object::Dictionary(dictionary),
                    ));
                }
            }
            "Page" => {}     // Ignored, processed later and separately
            "Outlines" => {} // Ignored, not supported yet
            "Outline" => {}  // Ignored, not supported yet
            _ => {
                merged_doc.max_id += 1;
                merged_doc.objects.insert(*object_id, object.clone());
            }
        }
    }

    for (object_id, object) in documents_pages.iter() {
        if let Ok(dictionary) = object.as_dict() {
            let mut dictionary = dictionary.clone();
            dictionary.set("Parent", pages_object.as_ref().unwrap().0);

            merged_doc
                .objects
                .insert(*object_id, Object::Dictionary(dictionary));
        }
    }

    let catalog_object = catalog_object.unwrap();
    let pages_object = pages_object.unwrap();

    // Build a new "Pages" with updated fields
    if let Ok(dictionary) = pages_object.1.as_dict() {
        let mut dictionary = dictionary.clone();

        // Set new pages count
        dictionary.set("Count", documents_pages.len() as u32);

        // Set new "Kids" list (collected from documents pages) for "Pages"
        dictionary.set(
            "Kids",
            documents_pages
                .into_keys()
                .map(Object::Reference)
                .collect::<Vec<_>>(),
        );

        merged_doc
            .objects
            .insert(pages_object.0, Object::Dictionary(dictionary));
    }

    // Build a new "Catalog" with updated fields
    if let Ok(dictionary) = catalog_object.1.as_dict() {
        let mut dictionary = dictionary.clone();
        dictionary.set("Pages", pages_object.0);
        dictionary.remove(b"Outlines"); // Outlines not supported in merged PDFs

        merged_doc
            .objects
            .insert(catalog_object.0, Object::Dictionary(dictionary));
    }

    merged_doc.trailer.set("Root", catalog_object.0);

    set_metadata(METADATA_TITLE, title, &mut merged_doc);
    set_metadata(METADATA_PRODUCER, PRODUCER, &mut merged_doc);

    // Update the max internal ID as wasn't updated before due to direct objects insertion
    merged_doc.max_id = merged_doc.objects.len() as u32;

    // Reorder all new Document objects
    merged_doc.renumber_objects();

    merged_doc.compress();
    merged_doc.save(out)?;

    Ok(())
}

/// Set metadata `key` to `value`.
///
/// Add the `Info trailer to the pdf document if it does not yet exist.`
fn set_metadata(key: &str, value: &str, doc: &mut Document) {
    let info_dict_id = match doc.trailer.get(b"Info") {
        Ok(&Object::Reference(id)) => id,
        _ => {
            // without this the following add_object call overwrites an existing
            // object at max_id
            doc.max_id += 1;

            let id = doc.add_object(Dictionary::new());
            doc.trailer.set("Info", Object::Reference(id));
            id
        }
    };

    if let Some(Object::Dictionary(ref mut info_dict)) = doc.objects.get_mut(&info_dict_id) {
        info_dict.set(key, Object::string_literal(value));
    }
}
